
FEATURE_NAMES = [
    "eigenvalue_sum",
    "omnivariance",
    "eigenentropy",
    "anisotropy",
    "planarity",
    "linearity",
    "PCA1",
    "PCA2",
    "surface_variation",
    "sphericity",
    "verticality",
    "nx",
    "ny",
    "nz",
]